#setencoding=utf8 
#
# Copyright (C) 2012 Sergey Metelev 
#
# version: v0.05-1-gaacc511 
#
# The following terms apply to usbcc.py and usbccpy.pdf files
#  
# The author hereby grant permission to use, copy, modify,
# distribute, and license this software and its documentation
# for any purpose, provided that existing copyright notices
# are retained in all copies and that this notice is included
# verbatim in any distributions. No written agreement,
# license, or royalty fee is required for any of the
# authorized uses.  Modifications to this software may be
# copyrighted by their authors and need not follow the
# licensing terms described here, provided that the new terms
# are clearly indicated on the first page of each file where
# they apply.
# 
# IN NO EVENT SHALL THE AUTHOR OR DISTRIBUTORS BE LIABLE TO
# ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
# CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
# SOFTWARE, ITS DOCUMENTATION, OR ANY DERIVATIVES THEREOF,
# EVEN IF THE AUTHOR HAVE BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# 
# THE AUTHOR AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE IS PROVIDED
# ON AN "AS IS" BASIS, AND THE AUTHOR AND DISTRIBUTORS HAVE
# NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
#setencoding=utf8
#
# В этом разделе содержится исходный текст программы с подробными
# комментариями. В конце раздела приводится список команд, который
# выполняется при автономном запуске модуля. Его можно менять, тем самым
# модуль можно рассматривать как автономное приложение для запуска
# простых последовательностей команд КАМАК.
#
# \subsection{Общее описание взаимодействия с адаптером}
# \subsubsection{Одиночные команды}
# Одиночные команды передаются адаптеру с помощью Setup-пакетов, тип
# передачи CONTROL (см. спецификацию
# USB \cite{usbstandard,chekunov200410,chekunov200412}). В
# модуле #pyusb# эта работа реализована с помощью функции
# #ctrl_transfer()#.  Параметры запроса соответствуют полям
# пакета Setup. В модуле pyusb функция определяется так:\\[3mm]
# # ctrl_transfer(bmRequestType, bRequest, wValue, wIndex,\#\\*\vspace{0mm}
# { }\hspace{20mm} # data_or_wLength, timeout):#\\[3mm]
# \label{ctrl_transfer}
# Описание аргументов:\begin{description}
#     \item[#bmRequestType#] определяет направление потока данных, тип
#          запроса и получателя
#     \item[#bRequest#] определяет запрос
#     \item[#wValue=0#] смысл поля определяется запросом. По умолчанию 0.
#     \item[#wIndex=0#] смысл поля определяется запросом. По умолчанию 0.
#     \item[#data_or_wLength#] при записи содержит побайтную последовательность
#          передаваемых данных, например:
#          #bytearray((0,1,2,3,4,5,6,7))#. В этом случае процедура
#          возвращает количество записанных байт.
#          При чтении содержит количество байт, которые необходимо
#          прочитать, а процедура возвращает прочитанные данные, тип
#          array.
#     \item[#timeout=100#] по умолчанию таймаут 100 миллисекунд.
# \end{description}
# Поле, #bmRequestType# конструируется следующим образом \cite{chekunov200412}:
#\begin{description}
#    \item[#D7#]  направление передачи данных в фазе данных; 0 OUT,
#                "от компьютера", 1 IN, "в компьютер"
#    \item[#D6, D5#]  тип требования: 0 стандартное, 1 класса,
#                 2 продавца (производителя), 3 зарезервировано;
#    \item[#D4...D0#]  получатель данного требования: 0 устройство,
#                 1 интерфейс, 2 точка, 3 остальные, 4...31 зарезервировано
#\end{description}
# Остальные поля, и в случае необходимости данные для передачи,
# формируются в соответствии с требованием.
# \subsubsection{Последовательности команд}
# Последовательности передаются с помощью передач типа BULK
# \cite{usbstandard,chekunov200410}. Функции, осуществляющие чтение и
# запись, очень похожи друг на друга. \\
# #device.write(OutEndPt, sequence, Interface, timeout)#\\
# #device.read(InEndPt, length, Interface, timeout)#\\
# Оба вызова содержат в качестве параметров #EndPt#---номер точки
# доступа, и необязательные #Interface#---номер интерфейся и
# #timeout#---ожидание ответа, в миллисекундах. Команда #write#
# дополнительно содержит #sequence#, байтовую последовательность для
# записи, и возвращает количество записанных байт. Команда #read#
# содержит #length#---количество байт для чтения и возвращает
# прочитанные данные. Формат данных определяется
# прошивкой. Он подробно изложен в аннотациях к функциям
# \hyperref[encode]{encode()} и \hyperref[decode]{decode()} 
#
# \subsection{Подготовительная часть}
# \subsubsection{Инициализация ведения лог-файла}
import logging
log=logging.getLogger('manager.usb')

# при автономном запуске добавляем обработчик сообщений и устанавливаем
# уровень для выводимых сообещний DEBUG
if __name__=="__main__":
    import logging.handlers
    common_format=  '%(levelname)-6s%(asctime)s %(message)s'
    date_format=    '%H:%M:%S'
    fmt=logging.Formatter(fmt=common_format, datefmt=date_format)
    # При автономной работе модуля устанавливаем уровень подробности
    # вывода сообщений и обработчика и логгера в DEBUG
    log.setLevel(logging.DEBUG)
    hand=logging.StreamHandler()
    hand.setFormatter(fmt)
    hand.setLevel(logging.DEBUG)
    log.addHandler(hand)

# \subsubsection{Инициализация констант, загрузка модулей}
# Именованные поля для конструирования типа запроса. На основе CyAPI.h,
# Cypress USB библиотеки под виндоус.\\
TGT_DEVICE, TGT_INTFC, TGT_ENDPT, TGT_OTHER = 0, 1, 2, 3 
REQ_STAD, REQ_CLASS, REQ_VENDOR = 0, 32, 64  
DIR_TO_DEVICE, DIR_FROM_DEVICE = 0, 128
    # Камак константы. Бит запрета, очистки, инициализации крейта и
    # программного сброса контроллера.
CAM_I, CAM_C, CAM_Z, CAM_Ax4 = 0x04, 0x02, 0x01, 0x40
    # Ожидание, при работе с последовательностью команд 
NoWait, WaitForQ, WaitForL = 0, 1, 128

import sys, os, time, usb

# \subsubsection{Поиск устройства}
# Осуществляется поиск устройства. Программа сообщает о найденном
# адаптере без firmware или с firmware. Если адаптер не найден или
# найден адаптер без прошивки, программа завершает работу. 

cypress_non_init = usb.core.find(idVendor=0x04b4, idProduct=0x8613)
if cypress_non_init is not None:
    log.error(u"Найден неинициализированный aдаптер usb-cc217.10")
    log.error(u"Добавьте в систему автоматическую инициализацию.")
    sys.exit()

cypress = usb.core.find(idVendor=0x0547, idProduct=0x1002)
if cypress is None:
    log.error(u"Не найден адаптер usb-cc217.10")
    sys.exit()
else:    
    log.info(u"Найден инициализированный адаптер usb-cc217.10")
    try:
        # Документация к pyusb (API 1.0) гласит, что при наличии одной
        # конфигурации указывать её не обязательно, но функцию установки
        # конфигурации надо вызывать. В то же время, при наличии одного
        # интерфейса можно даже не вызывать функцию, которая его
        # устанавливает. \cite{pyusbtutorial} Однако при этом повторные
        # загрузки модуля не сообщают об ошибке Resource Busy, то есть
        # возможна ситуация, когда два экземпляра программы управления
        # будут конкурировать за модуль. Поэтому сохраняем установку
        # интерфейса и альтернативного интерфейса.
        cypress.set_configuration(1)
        cypress.set_interface_altsetting(interface = 0, alternate_setting = 0)
    except usb.core.USBError, msg:
        log.error(u"Произошло исключение! причина: %s" % msg)
        sys.exit()


# \subsection{VendRquToDevice, VendRquFromDevice, VendRqu}
# Подробности по аргументам в описании функции, осуществляющей основную
# работу: \hyperref[ctrl_transfer]{ctrl\_transfer()}
def VendRquToDevice(bRequest, wValue=0, wIndex=0, data=bytearray((0,)), timeout=100):
    """ Пользовательский запрос с передачей данных к устройству.

    Аргументы: data: данные для записи (array)
    Возвращает: количество записанных байт    
    """
    rqType = TGT_DEVICE | REQ_VENDOR | DIR_TO_DEVICE   #0x40
    return cypress.ctrl_transfer(\
           rqType, bRequest, wValue, wIndex, data, timeout)

def VendRquFromDevice(bRequest, wValue=0, wIndex=0, wLength=0, timeout=100):
    """ Пользовательский запрос c передачей данных от устройства.

    Аргументы:
        wLength: кол-во байт для чтения (целое)
    Возвращает: прочитанные байты
    """
    rqType = TGT_DEVICE | REQ_VENDOR | DIR_FROM_DEVICE   #0xC0
    return cypress.ctrl_transfer(\
           rqType, bRequest, wValue, wIndex, wLength, timeout)

VendRqu=VendRquToDevice # Более короткая форма для записи пользовательского запроса к устройству.

# \subsection{encode(D, A, F, N, WaitFor), подготовка для загрузки в адаптер}
# \label{encode}
# Аргументы:
# #D# целое в диапазоне от 0 до $2^{24}-1=16777215$, #A#
# субадрес, #F# функция, #N# номер станции. #WaitFor#
# может быть одной из констант: #NoWait# = 0, сразу приступить к
# выполнению следующей команды; #WaitForQ# = 1 команда повторяется,
# пока не поступит подтверждение по Q,
# #WaitForL# = 128 ожидание LAM.
#
# Функция раскладывает #D# на три байта, Возвращает подготовленный для
# загрузки в адаптер набор:
# #(Hi, Md, Lo, A, F, N, 0, WaitFor)#, где #Hi# старший байт данных,
# #Md# средний байт данных, #Lo# младший байт данных, 7-й байт
# всегда ноль (незначащий), для остальных полей берутся аргументы
# функции без именений.
#
# Примечание: Если #D#>16777215, то разложение всё равно
# происходит и выводится предупреждение.

def encode(D, A, F, N, WaitFor):
    """Подготовка записей для загрузки в адаптер."""
    if D > 1<<24-1:
        log.warning(u"Превышен максимально допустимый размер данных для записи")
    (DD, Lo) = divmod(D , 256)
    (DD, Md) = divmod(DD, 256)
    (DD, Hi) = divmod(DD, 256)
    return (Hi, Md, Lo, A, F, N, 0, WaitFor)

# \subsection{decode(sequence, i=-1), расшифровка данных адаптера}
# \label{decode}
# Адаптер по каждой по каждой выполненной команде даёт
# информацию:\linebreak[4]
# #(A, F, N, LXQ, Hi, Md, Lo, S)# (содержит 8 элементов, каждый размером
# 1 байт). Здесь #A#
# субадрес, #F# функция, #N# номер станции, #LXQ# регистр
# статуса, который включает битовые поля:
# \begin{description}
#     \item[#7#] L=0 если есть хоть один LAM-сигнал (1 бит)
#     \item[#2..6#] LAM номер станции при наличии LAM-сигнала (5 бит)
#              (LAM=``look at me'', требование каких-то действий)
#     \item[#1#] X=0 если команда не прошла по каким-либо причинам (1 бит)
#     \item[#0#] Q=1 станция выполнила команду (1 бит)
# \end{description}
# Далее, #Hi# старший байт прочитанных данных, #Md# средний
# байт, #Lo# младший байт. #S#  была ли команда завершена или
# прекращена по таймауту. Если младший бит #S# равен 1, то было успешное
# завершение.
#
# Процедура decode() принимает последовательность, склеенную из
# произвольного количества вышеописанных записей и каждую преобразует в
# запись вида #(A, F, N, L, LAM, X, Q, S, D)#, (содержит 9
# элементов). Часть переменных повторяет входные аргументы, смысл переменных
# #L, LAM, X, Q# должен быть понятен из описания битовых полей
# регистра #LXQ#, данные собираются из байтов в число по формуле 
# #D=(Hi*256+Md)*256+Lo#
#
# Необязательный парамер #i# служит для возможности добавления номера
# итерации в лог-файл.
def decode(sequence, i=-1):
    """Преобразование результатов выполнения команд"""
    log.debug(u"чтение %3i, A,  F,  N,  L,LAM,  X,  Q,  S,  данные" % i)
    chunks=[]
    for i in range(len(sequence)/8):
        chunks.append(tuple(sequence[i*8:(i+1)*8]))
    result=[]
    for (A, F, N, LXQ, Hi, Md, Lo, S) in chunks:
        L, LAM, X, Q = LXQ>>7&1, LXQ>>2&int('11111',2), LXQ>>1&1, LXQ&1
        res_tuple = (A, F, N, L, LAM, X, Q, S&1, (Hi*256+Md)*256+Lo)
        log.debug(\
            u"           %2i, %2i, %2i, %2i, %2i, %2i, %2i, %2i, %i"\
            % res_tuple)
        result += [res_tuple,]
    return result

# \subsection{sequencer, класс для работы с последовательностями команд}
# В покомандном режиме каждый цикл обращения к КАМАК контроллеру и
# получения ответа от него занимает до 500мкс, это
# ограничение накладывает стандарт USB. В то же время адаптер способен
# работать с гораздо большей скоростью. Секвенсор даёт возможность
# загрузить последовательность команд и прочитать последовательность
# результатов их выполнения.
class sequencer(object):
    def __init__(self, sequence):
        self.InEndPt, self.OutEndPt  = 0x86, 0x02
        self.sequence=sequence
        self.write(sequence)

    # \subsubsection{write(self, sequence), Загрузка секвенсора}
    # Формат записей для секвенсора см. в функции \hyperref[encode]{encode()}
    # Для повторной загрузки последовательности нужна инициализация
    # адаптера (отключение и
    # подключение к шине USB). Без этого попытка повторной загрузки приводит к
    # таймауту. Поэтому при выходе по таймауту выдается предупреждение,
    # в котором рекомендуется перезагрузить адаптер.
    #
    # Первое чтение результатов возвращает саму загруженную последовательность.
    # Проверяем её на совпадение с той, которая была загружена.
    def write(self, sequence):
        """ Загрузка списка КАМАК-команд в адаптер.
        Аргументы:  (Hi, Md, Lo, A, F, N, 0, WaitFor) * N
        Возвращает: ---
        """
        VendRqu(0xB3)
        if (len(sequence)/8)*8 !=len(sequence):
            log.error(u"Число байт в последовательности не кратно 8!")
            sys.exit()
        else:
            log.debug(u"загружается %i команд" % (len(sequence)/8,))
        self.length=len(sequence)
        try:
            cypress.write(self.OutEndPt, sequence, 0, timeout=1000) # загрузка
        except usb.core.USBError, msg:
            log.error(u"Произошло исключение! причина: %s" % msg)
            if msg[0]==u"Operation timed out":
                log.error(u"Для повторной загрузки последовательности")
                log.error(u"отключите и вновь подключите адаптер к шине USB!")
            sys.exit()
        VendRqu(0xB6)
        cmp_sequence = cypress.read(self.InEndPt, self.length)
        if bytearray(sequence) != bytearray(cmp_sequence):
            log.error(u"Загруженная и прочитанная последовательности не совпадают!")
            sys.exit()
        self.start()    

    # \subsubsection{start(self), запуск последовательности}
    def start(self): VendRqu(0xD1)

    # \subsubsection{stop(self), останов работы последовательности}
    def stop(self): VendRqu(0xD2)   


    # \subsubsection{read(self), чтение результатов}
    # Возвращает последовательность результатов выполения команд.    
    # Формат записей см. в документации к функции \hyperref[decode]{decode()}:
    def read(self):
        """
        Аргументы: ---
        Возвращает (A, F, N, LXQ, Hi, Md, Lo, S) * N
        подробное описание в документации функции decode()
        """
        return cypress.read(self.InEndPt, self.length)


# \subsubsection{Camac(N, A, F, D), выполнение одной команды}
# Подробное описание аргументов и результата в функциях
# \hyperref[encode]{encode()} и \hyperref[decode]{decode()}
def Camac(N, A, F, D=0):
    """ Функция выполняет одну команду КАМАК.
    Аргументы: N, A, F, D
    Результат: A, F, N, LXQ, Hi, Md, Lo, S
    Подробное описание в документации функций encode() и decode()
    """
    VendRquToDevice(0xB7, data=encode(D, A, F, N, 0))
    return VendRquFromDevice(0xB9, wLength=8)

# \subsection{Функции инициализации крейта}
def CamSetZ():
    """Инициализация, установка бита Z"""
    VendRqu(0xB2, data=bytearray((CAM_Z,))) # 0x01

def CamSetC():
    """Очистка, установка бита C"""
    VendRqu(0xB2, data=bytearray((CAM_C,))) # 0x02

def CamSetI():
    """Запрет, установка бита I"""
    VendRqu(0xB2, data=bytearray((CAM_I,))) # 0x04

def CamReset():
    """Программный сброс контроллера, установка бита Ax4"""
    VendRqu(0xB2, data=bytearray((CAM_Ax4,))) # 0x40

# \subsubsection{Выполнение инициализации}
CamReset()
CamSetZ()


if __name__=="__main__":
    # \subsection{Команды, выполняемые при автономном запуске}
    # \label{autonome}
    # В случае, если файл запущен как программа, выполняем команды
    decode(Camac(0,0,0,0))
    decode(Camac(22,0,0,1))

    sequence=(\
              encode(0,  3, 25, 8, NoWait  ) +\
              encode(0, 12,  1, 8, WaitForQ) +\
              encode(0,  4,  0, 8, WaitForQ)\
             )
    s=sequencer(sequence)
    for i in range(3): decode(s.read(), i)

log.info(u"Загружен модуль usbcc.py")
